//! Implements writer that sends data to UE2 game server.

use std::cmp::{max, min};
use std::collections::VecDeque;
use std::convert::TryFrom;
use std::iter::Extend;

//  Maximum amount of bytes ue-server is able to receive at once
const UE_INPUT_BUFFER: usize = 4095;

/// For converting text messages into chunks of bytes that can be sent to the ue-server.
///
/// Every string message is converted into a length-prefixed array of utf8 bytes:
///
/// | Data | Length |
/// |---------|---------|
/// | Message `LENGTH` | 4 bytes: u32 BE |
/// | UTF8-encoded string | `LENGTH` bytes|
///
/// Resulting byte sequences from all messages are then concatenated (in the same order as they were
/// "written") into a single data stream. Bytes from the data stream are returned in chunks of
/// size no more than `UE_INPUT_BUFFER`. New chunk is returned only when `MessageWriter` knows
/// that ue-server's buffer has enough space to accept it.
///
/// Use `push()` to input string messages and `try_pop()` to retrieve next chunk.
/// NOTE: `try_pop()` can return `None` even if not all message data has been returned,
/// in case `MessageWriter` thinks that ue-server's buffer does not have enough space.
///
/// Call `update_ue_received_bytes()` to update `MessageWriter`'s information about
/// how many bytes ue-server has received so far. This can signal that its buffer has enough
/// free space. `MessageWriter` assumes that all data returned by its `try_pop()` method is sent
/// to ue-server.
///
/// Use `is_empty()` method to check for whether `MessageWriter` still has some data to return
/// (possibly after ``update_ue_received_bytes()`).
pub struct MessageWriter {
    sent_bytes: u64,
    ue_received_bytes: u64,
    pending_data: VecDeque<u8>,
}

impl MessageWriter {
    pub fn new() -> MessageWriter {
        MessageWriter {
            sent_bytes: 0,
            ue_received_bytes: 0,
            //  This value should be more than enough for typical use:
            //  it will take at least one second to send this much data to a 30 tick rate ue-server.
            pending_data: VecDeque::with_capacity(30 * UE_INPUT_BUFFER),
        }
    }

    pub fn push(&mut self, message: &str) {
        let message_as_utf8 = message.as_bytes();
        let message_as_utf8 = [
            &(message_as_utf8.len() as u32).to_be_bytes(),
            message_as_utf8,
        ]
        .concat();
        self.pending_data.extend(message_as_utf8.into_iter());
    }

    pub fn try_pop(&mut self) -> Option<Vec<u8>> {
        if self.is_empty() {
            return None;
        }
        let chunk_size = min(self.available_ue_buffer_capacity(), self.pending_data.len());
        if chunk_size == 0 {
            return None;
        }
        let mut bytes_to_send = Vec::with_capacity(chunk_size);
        for next_byte in self.pending_data.drain(..chunk_size) {
            bytes_to_send.push(next_byte);
        }
        self.sent_bytes += bytes_to_send.len() as u64;
        Some(bytes_to_send)
    }

    pub fn is_empty(&self) -> bool {
        self.pending_data.is_empty()
    }

    /// Takes total amount of bytes received so far by the ue-server, not just bytes received after
    /// the last `update_ue_received_bytes()` call.
    pub fn update_ue_received_bytes(&mut self, ue_received_bytes: u64) {
        self.ue_received_bytes = max(ue_received_bytes, self.ue_received_bytes);
    }

    fn available_ue_buffer_capacity(&self) -> usize {
        match usize::try_from(self.sent_bytes - self.ue_received_bytes).ok() {
            Some(ue_buffered_bytes) => max(0, UE_INPUT_BUFFER - ue_buffered_bytes),
            _ => 0,
        }
    }
}

#[test]
fn writer_contents_after_creation() {
    let writer = MessageWriter::new();
    assert_eq!(writer.is_empty(), true);
    assert_eq!(writer.available_ue_buffer_capacity(), UE_INPUT_BUFFER);
}

#[test]
fn writer_content_after_push() {
    let mut writer = MessageWriter::new();
    writer.push("Hello, world!");
    assert_eq!(writer.is_empty(), false);
    assert_eq!(writer.available_ue_buffer_capacity(), UE_INPUT_BUFFER);
}

#[test]
fn writing_single_short_message() {
    let mut writer = MessageWriter::new();
    writer.push("Hello, world!");
    let resulting_bytes = writer.try_pop().unwrap();
    let expected_bytes = [
        0, 0, 0, 13, // Bytes in the message
        b'H', b'e', b'l', b'l', b'o', b',', b' ', b'w', b'o', b'r', b'l', b'd', b'!',
    ];
    assert_eq!(writer.is_empty(), true);
    assert_eq!(
        writer.available_ue_buffer_capacity(),
        UE_INPUT_BUFFER - "Hello, world!".len() - 4
    );
    assert_eq!(resulting_bytes, expected_bytes);
    assert_eq!(writer.sent_bytes, expected_bytes.len() as u64);
}

#[test]
fn writing_first_chunk_of_single_long_message() {
    let mut writer = MessageWriter::new();
    //  Because we also have to pass message length, this will go over the sending limit
    let long_message = "Q".repeat(UE_INPUT_BUFFER);
    writer.push(&long_message);
    let resulting_bytes = writer.try_pop().unwrap();
    assert_eq!(writer.is_empty(), false);
    assert_eq!(resulting_bytes.len(), UE_INPUT_BUFFER);
    assert_eq!(writer.available_ue_buffer_capacity(), 0);
    //  Bytes in message = 4095 = 0x0fff
    assert_eq!(resulting_bytes[0], 0);
    assert_eq!(resulting_bytes[1], 0);
    assert_eq!(resulting_bytes[2], 0x0f);
    assert_eq!(resulting_bytes[3], 0xff);
    for &byte in resulting_bytes[4..].iter() {
        assert_eq!(byte, b'Q');
    }
    assert_eq!(writer.try_pop(), None);
    assert_eq!(writer.is_empty(), false);
}

#[test]
fn writing_second_chunk_of_single_long_message() {
    let mut writer = MessageWriter::new();
    //  Because we also have to pass lengths, this will go over the sending limit
    let long_message = "Q".repeat(UE_INPUT_BUFFER);
    writer.push(&long_message);
    //  This pops all but 4 bytes of `long_message`, that were required to encode message length
    let first_bytes = writer.try_pop().unwrap();
    writer.update_ue_received_bytes(first_bytes.len() as u64);
    let resulting_bytes = writer.try_pop().unwrap();
    assert_eq!(
        writer.available_ue_buffer_capacity(),
        UE_INPUT_BUFFER - resulting_bytes.len()
    );
    assert_eq!(writer.is_empty(), true);
    //  Bytes left for the next chunk = 4
    assert_eq!(resulting_bytes, [b'Q', b'Q', b'Q', b'Q'])
}
